"""
Musicbrainz related functions.
"""
import time
import musicbrainzngs as mus

from .__init__ import __version__


def init():
    """Initialize musicbrainz."""
    mus.set_useragent("python-pidi: A cover art daemon.",
                      __version__,
                      "https://github.com/pimoroni/pidi")


def get_cover(song, size=250, retry_delay=5, retries=5):
    """Download the cover art."""
    artist = song.get('artist')
    title = song.get('title')
    album = song.get('album', title)
    try:
        data = mus.search_releases(artist=artist,
                                   release=album,
                                   limit=1)
        release_id = data["release-list"][0]["release-group"]["id"]
        print("brainz: Using release-id: {}".format(data['release-list'][0]['id']))

        return mus.get_release_group_image_front(release_id, size=size)

    except mus.NetworkError as exc:
        if retries == 0:
            raise mus.NetworkError("Failure connecting to MusicBrainz.org") from exc
        print("brainz: Warning, retrying download. {retries} retries left!".format(retries=retries))
        time.sleep(retry_delay)
        get_cover(song, size, retries=retries - 1)

    except mus.ResponseError:
        print("brainz: Error, couldn't find album art for",
              "{artist} - {album}".format(artist=artist, album=album))

    return None



"""
Get song info.
"""
import xml
import shutil
from base64 import decodebytes
from pkg_resources import iter_entry_points

import mpd
import untangle
from .fifo import FIFO

from . import brainz
from . import util


def get_client_types():
    """Enumerate the pidi.plugin.client entry point and return installed client types."""
    client_types = {
        'mpd': ClientMPD,
        'ssnc': ClientShairportSync
    }

    for entry_point in iter_entry_points("pidi.plugin.client"):
        try:
            plugin = entry_point.load()
            client_types[plugin.option_name] = plugin
        except (ModuleNotFoundError, ImportError) as err:
            print("Error loading client plugin {entry_point}: {err}".format(
                entry_point=entry_point,
                err=err
            ))

    return client_types


class ClientShairportSync():
    """Client for ShairportSync metadata pipe."""
    # pylint: disable=too-many-instance-attributes
    def __init__(self, args):
        self.title = ""
        self.artist = ""
        self.album = ""
        self.time = 100
        self.state = ""
        self.volume = 0
        self.random = 0
        self.repeat = 0
        self.shuffle = 0
        self.album_art = ""
        self.pending_art = False

        self._update_pending = False

        self.fifo = FIFO(args.pipe, eol="</item>", skip_create=True)

    def add_args(argparse):  # pylint: disable=no-self-argument
        """Expand argparse instance with client-specific args."""
        argparse.add_argument(
            "--pipe",
            help="Pipe file for shairport sync metadata.",
            default="/tmp/shairport-sync-metadata")

    def status(self):
        """Return current status details."""
        return {
            "random": self.random,
            "repeat": self.repeat,
            "state": self.state,
            "volume": self.volume,
            "shuffle": self.shuffle
        }

    def currentsong(self):
        """Return current song details."""
        self._update_pending = False
        return {
            "title": self.title,
            "artist": self.artist,
            "album": self.album,
            "time": self.time
        }

    def get_art(self, cache_dir, size):  # pylint: disable=unused-argument
        """Get the album art."""
        if self.album_art == "" or self.album_art is None:
            util.bytes_to_file(util.default_album_art(), cache_dir / "current.jpg")
            return

        util.bytes_to_file(self.album_art, cache_dir / "current.jpg")

        self.pending_art = False

    def update_pending(self):
        """Check if a new update is pending."""
        attempts = 0
        while True:
            data = self.fifo.read()
            if data is None or len(data) == 0:
                attempts += 1
                if attempts > 100:
                    return False
            else:
                self._parse_data(data)
                self._update_pending = True

        return self._update_pending

    def _parse_data(self, data):
        try:
            data = untangle.parse(data)
        except (xml.sax.SAXException, AttributeError) as exp:
            print(f"ClientShairportSync: failed to parse XML ({exp})")
            return

        dtype = bytes.fromhex(data.item.type.cdata).decode("ascii")
        dcode = bytes.fromhex(data.item.code.cdata).decode("ascii")

        data = getattr(data.item, "data", None)

        if data is not None:
            encoding = data["encoding"]
            data = data.cdata
            if encoding == "base64":
                data = decodebytes(data.encode("ascii"))

        if (dtype, dcode) == ("ssnc", "PICT"):
            self.pending_art = True
            self.album_art = data

        if (dtype, dcode) == ("core", "asal"):  # Album
            self.album = "" if data is None else data.decode("utf-8")

        if (dtype, dcode) == ("core", "asar"):  # Artist
            self.artist = "" if data is None else data.decode("utf-8")

        if (dtype, dcode) == ("core", "minm"):  # Song Name / Item
            self.title = "" if data is None else data.decode("utf-8")

        if (dtype, dcode) == ("ssnc", "prsm"):
            self.state = "play"

        if (dtype, dcode) == ("ssnc", "pend"):
            self.state = "stop"


class ClientMPD():
    """Client for MPD and MPD-like (such as Mopidy) music back-ends."""
    def __init__(self, args=None):
        """Initialize mpd."""
        self._client = mpd.MPDClient()
        self._current = None

        try:
            print(f"Connecting to mpd {args.server}:{args.port}")
            self._client.connect(args.server, args.port)
            print("Connected!")

        except ConnectionRefusedError as exc:
            raise RuntimeError("error: Connection refused to mpd/mopidy.") from exc

    def add_args(argparse):  # pylint: disable=no-self-argument
        """Expand argparse instance with client-specific args."""
        argparse.add_argument(
            "--port",
            help="Use a custom mpd port.",
            default=6600)

        argparse.add_argument(
            "--server",
            help="Use a remote server instead of localhost.",
            default="localhost")

    def currentsong(self):
        """Return current song details."""
        result = self._client.currentsong()  # pylint: disable=no-member
        return result

    def status(self):
        """Return current status details."""
        result = self._client.status()  # pylint: disable=no-member
        return result

    def update_pending(self, timeout=0.1):  # pylint: disable=unused-argument,no-self-use
        """Determine if anything has changed on the server."""
        return False

    def get_art(self, cache_dir, size):
        """Get the album art."""
        song = self.currentsong()
        if len(song) < 2:
            print("mpd: Nothing currently playing.")
            util.bytes_to_file(util.default_album_art(), cache_dir / "current.jpg")
            return

        artist = song.get('artist')
        title = song.get('title')
        album = song.get('album', title)
        file_name = "{artist}_{album}_{size}.jpg".format(
            artist=artist,
            album=album,
            size=size
        ).replace("/", "")
        file_name = cache_dir / file_name

        if file_name.is_file():
            shutil.copy(file_name, cache_dir / "current.jpg")
            print("mpd: Found cached art.")

        else:
            print("mpd: Downloading album art...")

            brainz.init()
            album_art = brainz.get_cover(song, size)

            if not album_art:
                album_art = util.default_album_art()

            util.bytes_to_file(album_art, cache_dir / file_name)
            util.bytes_to_file(album_art, cache_dir / "current.jpg")

            print("mpd: Swapped art to {artist}, {title}.".format(
                artist=artist,
                title=title
            ))



# -*- coding: utf-8 -*-
"""
   ▄███████▄  ▄█  ████████▄   ▄█
  ███    ███ ███  ███   ▀███ ███
  ███    ███ ███▌ ███    ███ ███▌
  ███    ███ ███▌ ███    ███ ███▌
▀█████████▀  ███▌ ███    ███ ███▌
  ███        ███  ███    ███ ███
  ███        ███  ███   ▄███ ███
 ▄████▀      █▀   ████████▀  █▀

Pidi - Pirate Display

A fork of "bum" originally
created by Dylan Araps
"""
import argparse
import pathlib
import time
import sys

from .display import get_display_types
from .client import get_client_types

from .__init__ import __version__


def get_args(display_types, client_types):
    """Get the script arguments."""
    description = "pidi - Download and display album art \
                   for mpd tracks."
    arg = argparse.ArgumentParser("pidi", description=description)

    arg.add_argument("--update-interval", metavar="\"i\"",
                     help="overlay update interval in seconds.",
                     default=1.0,
                     type=float)

    arg.add_argument("--fps",
                     help="frames per second.",
                     type=int, default=30)

    arg.add_argument("--size", metavar="\"px\"",
                     help="what size to display the album art in.",
                     default=240)

    arg.add_argument("--cache_dir", metavar="\"/path/to/dir\"",
                     help="Where to store the downloaded cover art.",
                     default=pathlib.Path.home() / ".cache/pidi",
                     type=pathlib.Path)

    arg.add_argument("--version", action="store_true",
                     help="Print \"pidi\" version.")

    arg.add_argument("--no_display",
                     action="store_true",
                     help="Only download album art, don't display.")

    arg.add_argument("--display", choices=display_types.keys(),
                     help="Display class to use.",
                     default='dummy')

    arg.add_argument("--client", choices=client_types.keys(),
                     help="Client class to use.",
                     default='mpd')

    # Strip out --help so we can parse_known_args
    # without triggering help text output.
    has_help = False
    try:
        sys.argv.remove('--help')
        has_help = True
    except ValueError:
        pass

    args, _ = arg.parse_known_args()

    # Add display/client specific args into the parser
    display_types[args.display].add_args(arg)

    client_types[args.client].add_args(arg)

    # Add --help back if it was supplied
    if has_help:
        sys.argv.append('--help')

    return arg.parse_args()


def process_args(args):
    """Process the arguments."""
    if args.version:
        print("pidi {version}".format(version=__version__))
        sys.exit(0)


def main():
    """Main script function."""
    display_types = get_display_types()
    client_types = get_client_types()

    args = get_args(display_types, client_types)
    process_args(args)

    if args.no_display:
        print("Warning --no_display overrides --display option!")
        args.display = 'dummy'

    display = display_types[args.display](args)

    client = client_types[args.client](args)

    last_track = ''
    last_update = 0

    while True:
        if client.update_pending() or time.time() - last_update > float(args.update_interval):
            status = client.status()
            currentsong = client.currentsong()

            if status == {} or currentsong == {}:
                pass  # No status or song info available
            else:
                title = currentsong.get('title', 'Untitled')
                artist = currentsong.get('artist', 'No Artist')
                album = currentsong.get('album', title)

                current_track = f"{title} - {artist}, {album}"

                if current_track != last_track:
                    print(f"pidi: got new track: {current_track}")

                get_art = current_track != last_track or getattr(client, "pending_art", False)

                if get_art:
                    client.get_art(args.cache_dir, args.size)
                    display.update_album_art(args.cache_dir / "current.jpg")
                    last_track = current_track

                display.update_overlay(
                    status['random'] == '1',
                    status['repeat'] == '1',
                    status['state'],
                    int(status['volume']),
                    float(status.get('elapsed', 0)) / float(currentsong['time']),
                    float(status.get('elapsed', 0)),
                    title,
                    album,
                    artist
                )

                last_update = time.time()

        display.redraw()

        time.sleep(1.0 / args.fps)


if __name__ == "__main__":
    main()



"""
Basic fifo handler.
"""
import os
import select


class FIFO:
    """
    Basic fifo handler.
    """
    def __init__(self, fifo_name, eol="\n", skip_create=False):
        self.fifo_name = fifo_name
        self.eol = eol

        if not skip_create:
            self._create()

        self._f = None
        self._buf = ""

    def _create(self):
        try:
            os.unlink(self.fifo_name)
        except (IOError, OSError):
            pass

        os.mkfifo(self.fifo_name)
        os.chmod(self.fifo_name, 0o777)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self._f is not None:
            os.close(self._f)
            self._f = None

    def read(self):
        """Read data from the fifo."""
        if self._f is None:
            self._f = os.open(self.fifo_name, os.O_RDONLY | os.O_NONBLOCK)

        # Check for inbound command
        fifos, _, _ = select.select([self._f], [], [], 0)
        if self._f in fifos:
            while True:
                try:
                    char = os.read(self._f, 1).decode("UTF-8")
                except BlockingIOError:
                    return None
                if len(char) == 0:
                    return None
                self._buf += char
                if self._buf.endswith(self.eol):
                    buf = self._buf
                    self._buf = ""
                    return buf.strip()
        else:
            return None
